import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;

import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import javax.swing.*;
import java.util.Map;
import java.util.HashMap;
import java.util.Random;

import static java.lang.Math.*;

public class LSystem implements Runnable {
	static final float RADIUS = 0.5f;
	static final float RANGE  = 1.0f;
	static final BasicStroke STROKE = new BasicStroke(RADIUS / 3);
	static final Random random = new Random();

	static class Turtle {
		double minX;
		double maxX;
		double minY;
		double maxY;

		double x;
		double y;
		double direction;
		double step;

		Path2D.Double path;

		TurtleStack stack;

		void reset () {
			this.x = 0.0;
			this.y = 0.0;

			this.minX = 0.0;
			this.minY = 0.0;

			this.maxX = 0.0;
			this.maxY = 0.0;

			this.step = 100.0;

			this.direction = toRadians(direction);

			this.path = new Path2D.Double();

			path.moveTo(x, y);

			this.stack = null;
		}

		void foreward () {
			x += step * sin(direction);
			y -= step * cos(direction);

			path.lineTo(x, y);

			updateBounds();
		}

		void updateBounds () {
			minX = min(x, minX);
			minY = min(y, minY);
			maxX = max(x, maxX);
			maxY = max(y, maxY);
		}

		void go () {
			x += step * sin(direction);
			y -= step * cos(direction);

			path.lineTo(x, y);

			updateBounds();
		}

		void turn (int angle) {
			direction += toRadians(angle);
		}

		void push () {
			stack = new TurtleStack(x, y, step, direction, stack);
		}

		void pop () {
			x = stack.x;
			y = stack.y;
			step = stack.step;
			direction = stack.direction;
			step = stack.step;
			stack = stack.next;
			path.moveTo(x, y);
		}

		void process (String sequence, int angle) {
			reset();

			int count = 1;
			char prev = 0;

			for (char ch : sequence.toCharArray()) {
				switch (ch) {
					case '|': foreward(); step *= 0.65; break;
					case 'F': foreward(); break;
					case 'G': go(); break;
					case '+': turn(count *  angle); break;
					case '-': turn(count * -angle); break;
					case '[': push(); break;
					case ']': pop(); break;
				}

				if (ch >= '0' && ch <= '9') {
					count = ch - '0';
				} else {
					count = 1;
				}

				prev = ch;
			}
		}

		static class TurtleStack {
			final double x;
			final double y;
			final double step;
			final double direction;
			final TurtleStack next;

			TurtleStack (double x, double y, double step, double direction, TurtleStack next) {
				this.x = x;
				this.y = y;
				this.step = step;
				this.direction = direction;
				this.next = next;
			}
		}

	}

	int 				angle = 25;
	int				steps = 12;
	String 				axiom = "F";
	HashMap<Character, String> 	rules = new HashMap<Character, String>();

	public LSystem (String...args) {
		for (String arg : args) {
			if (arg.indexOf(':') > 0) {
				String[] parts = arg.split(":");

				if (parts.length != 2) continue;

				if ("-angle".equals(parts[0])) {
					angle = Integer.parseInt(parts[1]);
				}

				if ("-axiom".equals(parts[0])) {
					axiom = parts[1];
				}

				if ("-steps".equals(parts[0])) {
					steps = Integer.parseInt(parts[1]);
				}

			} else {
				String[] parts = arg.split("=");

				rules.put(parts[0].charAt(0), parts[1]);
			}
		}

		if (rules.isEmpty()) {
			rules.put('F', "|[-F][+F]");
		}
	}

	static final int PAD = 20;

	@Override
	@SuppressWarnings("unchecked")
	public void run () {
		final JFrame frame = new JFrame("LSystem");

		final Turtle turtle = new Turtle();

		turtle.process(axiom, angle);

		final JPanel panel = new JPanel () {
			final BasicStroke stroke = new BasicStroke(1.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);

			@Override
			public void paintComponent (Graphics g) {
				Graphics2D graphics = (Graphics2D)g;

				int width = getWidth();
				int height = getHeight();

				graphics.setColor(Color.WHITE);
				graphics.fillRect(0, 0, width, height);

				double cx = (turtle.maxX + turtle.minX) * 0.5;
				double cy = (turtle.maxY + turtle.minY) * 0.5;
				double wx = max(0.1, turtle.maxX - turtle.minX);
				double wy = max(0.1, turtle.maxY - turtle.minY);

				double scale = min((width - PAD)/ wx, (height - PAD) / wy);

				AffineTransform trans = AffineTransform.getTranslateInstance(width/2, height/2);

				trans.scale(scale, scale);
				trans.translate(-cx, -cy);

				graphics.setTransform(trans);
				graphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
				graphics.setColor(Color.BLACK);
				graphics.setStroke(stroke);

				graphics.draw(turtle.path);
			}
		};

		new Timer(750, new ActionListener() {
			String state = axiom;
			int count = 0;

			@Override
			public void actionPerformed (ActionEvent event) {
				if (count % (steps + 4) == 0) {
					state = axiom;
				} else if (count % (steps + 4) <= steps) {
					// create new state
					StringBuilder buffer = new StringBuilder();

					for (char ch : state.toCharArray()) {
						String expansion = rules.get(ch);

						if (expansion != null) {
							buffer.append(expansion);
						} else {
							buffer.append(ch);
						}
					}

					state = buffer.toString();
				}

				++count;

				System.out.println(state);

				turtle.process(state, angle);

				panel.repaint();
			}
		}).start();

		frame.setLayout(new BorderLayout());
		frame.add(panel, BorderLayout.CENTER);

		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		frame.setVisible(true);
		frame.setSize(800, 600);
	}

	public static void main (String...args) {
		SwingUtilities.invokeLater(new LSystem(args));
	}
}
